// This is the source code of the HPTTagger as used in 1308.0540
// If you use it please cite
//
// %\cite{Schaetzel:2013vka}
// \bibitem{Schaetzel:2013vka}
//  S.~Schaetzel and M.~Spannowsky,
//  %``Tagging highly boosted top quarks,''
//  Phys.\ Rev.\ D {\bf 89}, 014007 (2014)
//  [arXiv:1308.0540 [hep-ph]].
//  %%CITATION = ARXIV:1308.0540;%%
//
// and due to the similarities with the HEPTopTagger algorithm please consider citing
//
// %\cite{Plehn:2010st}
// \bibitem{Plehn:2010st}
//  T.~Plehn, M.~Spannowsky, M.~Takeuchi and D.~Zerwas,
//  %``Stop Reconstruction with Tagged Tops,''
//  JHEP {\bf 1010}, 078 (2010)
//  [arXiv:1006.2833 [hep-ph]].
//  %%CITATION = ARXIV:1006.2833;%%
//

// The first input argument of the HPTTagger is the ClusterSequence of the track/e-cal-based fat jet
// The second input argument is the calibrated corresponding (fat) jet
// The third input argument is the cluster sequence of the full fat jet based on hcal+ecal+tracks

#ifndef __HPTTOPTAGGER_HH__
#define __HPTTOPTAGGER_HH__

using namespace std;
using namespace fastjet;

class HPTTopTagger {
   public:

      HPTTopTagger(const fastjet::ClusterSequence & ccs,
            const fastjet::PseudoJet & cjet,
            const fastjet::ClusterSequence & cs,
            const fastjet::PseudoJet & jet);

      HPTTopTagger(const fastjet::ClusterSequence & ccs,
            const fastjet::PseudoJet & cjet,
            const fastjet::ClusterSequence & cs,
            const fastjet::PseudoJet & jet,
            double mtmass, double mwmass);

      void run_tagger();
      bool is_maybe_top() const {return _is_maybe_top;}
      bool is_masscut_passed() const {return _is_masscut_passed;}
      bool is_wmass_passed() const {return _is_wmass_passed;}
      const PseudoJet & top_candidate() const {return _top_candidate;}
      double ratalpha() const {return _alpha;}
      const PseudoJet & W_candidate() const {return _W_candidate;}
      const vector<PseudoJet> & top_subjets() const {return _top_subjets;}
      const vector<PseudoJet> & top_hadrons() const {return _top_hadrons;}
      unsigned top_count() const {return _top_count;}
      const vector<PseudoJet> & hardparts() const {return _top_parts;}
      unsigned parts_size() const {return _parts_size;}
      double delta_top() const {return _delta_top;}
      double drmin() const {return _drmin;}

      const vector<vector<PseudoJet> > & candjets() const {return _candjets;}
      void get_setting() const;
      void get_info() const;
      // for setting parameters
      void set_max_subjet_mass(double x) {_max_subjet_mass=x;}
      void set_mass_drop_threshold(double x) {_mass_drop_threshold=x;}
      void set_top_range(double xmin, double xmax) {_mtmin=xmin; _mtmax=xmax;}
      void set_mass_ratio_range(double rmin, double rmax) {_rmin=rmin; _rmax=rmax;}
      void set_mass_ratio_cut(double m23cut, double m13cutmin,double m13cutmax){_m23cut=m23cut; _m13cutmin=m13cutmin; _m13cutmax=m13cutmax;}

      void set_nfilt(unsigned nfilt) {_nfilt=nfilt;}
      void set_filtering_jetalgorithm(JetAlgorithm jet_algorithm) {_jet_algorithm=jet_algorithm;}
      void set_reclustering_jetalgorithm(JetAlgorithm jet_algorithm) {_jet_algorithm_recluster=jet_algorithm;}
      //
      double cos_theta_h() const;
      double dr_bjj() const;
      vector<double> dr_values() const;
      bool debugg;

   private:
      const ClusterSequence * _cs;
      const ClusterSequence * _ccs;
      const PseudoJet _jet;
      const PseudoJet _cjet;
      const double _mtmass, _mwmass;
      double _mass_drop_threshold;
      double _max_subjet_mass; // stop when subjet mass < 30 GeV
      double _mtmin, _mtmax;
      double _mwmin, _mwmax;
      double _rmin, _rmax;
      double _m23cut, _m13cutmin, _m13cutmax;
      size_t _nfilt;
      double _alpha;
      double _drmin;
      // filtering algorithm
      JetAlgorithm _jet_algorithm;
      JetAlgorithm _jet_algorithm_recluster;

      bool _is_masscut_passed;
      bool _is_wmass_passed;
      bool _is_maybe_top;
      double _delta_top;
      unsigned _top_count;
      unsigned _parts_size;
      PseudoJet _top_candidate;
      PseudoJet _W_candidate;
      vector<PseudoJet> _top_subjets;
      vector<PseudoJet> _top_hadrons;
      vector<PseudoJet> _top_parts;
      vector<PseudoJet> _top_charged_parts;
      vector<vector<PseudoJet> > _candjets;

      void FindHardSubst(const PseudoJet& jet, vector<fastjet::PseudoJet>& t_parts);
      vector<PseudoJet> Filtering(const vector <PseudoJet> & top_constits, const JetDefinition & filtering_def);
      void store_topsubjets(const vector<PseudoJet>& top_subs);
      bool check_mass_criteria(const vector<fastjet::PseudoJet> & top_subs) const;
      bool check_Wmass(const vector<fastjet::PseudoJet> & top_subs, fastjet::PseudoJet & Wcand) const;
      double check_cos_theta(const PseudoJet & jet, const PseudoJet & subj1,const PseudoJet & subj2) const;
      PseudoJet Sum(const vector<PseudoJet>& );
      double r_max_3jets(const fastjet::PseudoJet & jet1,const fastjet::PseudoJet & jet2,
            const fastjet::PseudoJet & jet3) const;

      double r_min_subjets(const vector<fastjet::PseudoJet> & jets) const;
};

//--------------------------------------------------------------------
double HPTTopTagger::cos_theta_h() const {
   return check_cos_theta(_top_candidate,_top_subjets[1],_top_subjets[2]);// m23 is closest to mW
}

double HPTTopTagger::dr_bjj() const{
   if(_top_subjets.size()!=3){return -1;}
   return r_max_3jets(_top_subjets[0],_top_subjets[1],_top_subjets[2]);
}

vector<double> HPTTopTagger::dr_values() const{
   vector<double> dr_values;
   dr_values.push_back(sqrt(_top_subjets[1].squared_distance(_top_subjets[2])));
   dr_values.push_back(sqrt(_top_subjets[0].squared_distance(_top_subjets[2])));
   dr_values.push_back(sqrt(_top_subjets[0].squared_distance(_top_subjets[1])));
   return dr_values;
}


double HPTTopTagger::r_min_subjets(const vector<fastjet::PseudoJet> & jets) const{

   double drmin(10000);
   for(unsigned ii=0;ii<jets.size();ii++)
   {
      for(unsigned jj=ii+1; jj<jets.size();jj++)
      {
         if(sqrt(jets[ii].squared_distance(jets[jj])) < drmin) drmin = sqrt(jets[ii].squared_distance(jets[jj]));
      }
   }

   return(drmin);
}


double HPTTopTagger::r_max_3jets(const fastjet::PseudoJet & jet1,const fastjet::PseudoJet & jet2,const fastjet::PseudoJet & jet3) const{
   fastjet::PseudoJet jet12,jet13,jet23;
   jet12=jet1+jet2;
   jet13=jet1+jet3;
   jet23=jet2+jet3;

   double a=sqrt(jet1.squared_distance(jet2));
   double b=sqrt(jet2.squared_distance(jet3));
   double c=sqrt(jet3.squared_distance(jet1));
   double dR1=a,dR2=a;

   if(a<=b && a<=c){
      dR1=a;
      dR2=sqrt(jet12.squared_distance(jet3));
   };
   if(b<a && b<=c){
      dR1=b;
      dR2=sqrt(jet23.squared_distance(jet1));
   };
   if(c<a && c<b){
      dR1=c;
      dR2=sqrt(jet13.squared_distance(jet2));
   };
   return max(dR1,dR2);
}

double HPTTopTagger::check_cos_theta(const PseudoJet & jet,const PseudoJet & subj1,const PseudoJet & subj2) const
{
   // the two jets of interest: top and lower-pt prong of W
   PseudoJet W2;
   PseudoJet top = jet;

   if(subj1.perp2() < subj2.perp2())
   {
      W2 = subj1;
   }
   else
   {
      W2 = subj2;
   }

   // transform these jets into jets in the rest frame of the W
   W2.unboost(subj1+subj2);
   top.unboost(subj1+subj2);

   double csthet = (W2.px()*top.px() + W2.py()*top.py() + W2.pz()*top.pz())/sqrt(W2.modp2() * top.modp2());
   return(csthet);
}

void HPTTopTagger::FindHardSubst(const PseudoJet & this_jet, vector<fastjet::PseudoJet> & t_parts)
{

   PseudoJet parent1(0,0,0,0), parent2(0,0,0,0);
   if (this_jet.m() < _max_subjet_mass || !this_jet.validated_cs()->has_parents(this_jet, parent1, parent2))
   {
      t_parts.push_back(this_jet);
   }
   else
   {
      if (parent1.m() < parent2.m()) swap(parent1, parent2);

      FindHardSubst(parent1,t_parts);

      if (parent1.m() < _mass_drop_threshold * this_jet.m())
      {
         FindHardSubst(parent2,t_parts);
      }
   }

}

void HPTTopTagger::store_topsubjets(const vector<PseudoJet>& top_subs){
   _top_subjets.resize(0);
   double m12=(top_subs[0]+top_subs[1]).m();
   double m13=(top_subs[0]+top_subs[2]).m();
   double m23=(top_subs[1]+top_subs[2]).m();
   double dm12=abs(m12-_mwmass);
   double dm13=abs(m13-_mwmass);
   double dm23=abs(m23-_mwmass);
   if(dm23<=dm12 && dm23<=dm13){
      _top_subjets.push_back(top_subs[0]); //supposed to be b
      _top_subjets.push_back(top_subs[1]); //W-jet 1
      _top_subjets.push_back(top_subs[2]); //W-jet 2
   }
   else if(dm13<=dm12 && dm13<dm23){
      _top_subjets.push_back(top_subs[1]); //supposed to be b
      _top_subjets.push_back(top_subs[0]); //W-jet 1
      _top_subjets.push_back(top_subs[2]); //W-jet 2
   }
   else if(dm12<dm23 && dm12<dm13){
      _top_subjets.push_back(top_subs[2]); //supposed to be b
      _top_subjets.push_back(top_subs[0]); //W-jet 1
      _top_subjets.push_back(top_subs[1]); //W-jet 2
   }
   return;
}

bool HPTTopTagger::check_mass_criteria(const vector<PseudoJet> & top_subs) const{
   bool is_passed=false;
   double m12=(top_subs[0]+top_subs[1]).m();
   double m13=(top_subs[0]+top_subs[2]).m();
   double m23=(top_subs[1]+top_subs[2]).m();
   double m123=(top_subs[0]+top_subs[1]+top_subs[2]).m();
   if(
         (atan(m13/m12)>_m13cutmin && _m13cutmax > atan(m13/m12)
          && (m23/m123>_rmin && _rmax>m23/m123))
         ||
         (((m23/m123)*(m23/m123) < 1-_rmin*_rmin*(1+(m13/m12)*(m13/m12))) &&
          ((m23/m123)*(m23/m123) > 1-_rmax*_rmax*(1+(m13/m12)*(m13/m12))) &&
          (m23/m123 > _m23cut))
         ||
         (((m23/m123)*(m23/m123) < 1-_rmin*_rmin*(1+(m12/m13)*(m12/m13))) &&
          ((m23/m123)*(m23/m123) > 1-_rmax*_rmax*(1+(m12/m13)*(m12/m13))) &&
          (m23/m123 > _m23cut))
     ){
      is_passed=true;
   }
   return is_passed;
}


bool HPTTopTagger::check_Wmass(const vector<PseudoJet> & top_subs, fastjet::PseudoJet & Wcand) const
{
   bool is_passed=false;
   double tmpdeltaW(100000.0);

   for(unsigned rr=0; rr<top_subs.size(); rr++)
   {
      for(unsigned ll=rr+1; ll<top_subs.size(); ll++)
      {

         double deltaW = abs((top_subs[rr]+top_subs[ll]).m() - _mwmass);
         if(deltaW < tmpdeltaW)
         {
            tmpdeltaW = deltaW;
            Wcand = top_subs[rr]+top_subs[ll];
         }

      }
   }

   if((Wcand.m()<_mwmax) && (Wcand.m()>_mwmin))
   {
      is_passed=true;
   }
   return is_passed;
}

////////// Top-TAGGER: /////////////////////////////////////////////////////////////////
HPTTopTagger::HPTTopTagger(const fastjet::ClusterSequence & ccs,
      const fastjet::PseudoJet & cjet,
      const fastjet::ClusterSequence & cs,
      const fastjet::PseudoJet & jet) :
   debugg(false),
   _cs(&cs), _ccs(&ccs), _jet(jet), _cjet(cjet), _mtmass(172.3), _mwmass(80.4),
   _mass_drop_threshold(0.8), _max_subjet_mass(20.),
   _mtmin(172.3 - 25.),_mtmax(172.3 + 25.), _mwmin(80.4-15.0),_mwmax(80.4+15.0),
   _rmin(0.85*80.4/172.3),_rmax(1.15*80.4/172.3),
   _m23cut(0.35),_m13cutmin(0.2),_m13cutmax(1.3),
   _nfilt(3),_jet_algorithm(cambridge_algorithm),_jet_algorithm_recluster(cambridge_algorithm)
{}

HPTTopTagger::HPTTopTagger(const fastjet::ClusterSequence & ccs,
      const fastjet::PseudoJet & cjet,
      const fastjet::ClusterSequence & cs,
      const fastjet::PseudoJet & jet,
      double mtmass,double mwmass) :
   debugg(false),
   _cs(&cs), _ccs(&ccs), _jet(jet), _cjet(cjet), _mtmass(mtmass), _mwmass(mwmass),
   _mass_drop_threshold(0.8), _max_subjet_mass(20.),
   _mtmin(mtmass - 25.),_mtmax(mtmass + 25.), _mwmin(mwmass-15.0),_mwmax(mwmass+15.0),
   _rmin(0.85*mwmass/mtmass),_rmax(1.15*mwmass/mtmass),
   _m23cut(0.35),_m13cutmin(0.2),_m13cutmax(1.3),
   _nfilt(3),_jet_algorithm(cambridge_algorithm),_jet_algorithm_recluster(cambridge_algorithm)
{}

void HPTTopTagger::run_tagger()
{
   _delta_top=1000000000000.0;
   _top_candidate.reset(0.,0.,0.,0.);
   _W_candidate.reset(0.,0.,0.,0.);
   _top_count=0;
   _parts_size=0;
   _is_maybe_top=false;
   _is_masscut_passed=false;
   _is_wmass_passed=false;
   _top_subjets.clear();
   _top_hadrons.clear();
   _top_parts.clear();
   _top_charged_parts.clear();
   _alpha=0;

   if(debugg)
   {
      cout << "mtmass in top_tagger: " << _mtmass << endl;
      cout << "mwmass in top_tagger: " << _mwmass << endl;
      cout << "jet input HPTTopTagger: " << endl;
      //printjet(_jet);
   }

   _alpha = _jet.e()/_cjet.e();
   // input this_jet, output _top_parts
   FindHardSubst(_cjet, _top_parts);

   if(_top_parts.size() < 2) return;

   _top_parts = sorted_by_pt(_top_parts);

   // store hard substructure of the top candidate
   _parts_size=_top_parts.size();
   _drmin =r_min_subjets(_top_parts);

   // collect all constituents from charged subjets
   vector <PseudoJet> top_constits;
   top_constits.clear();

   for(unsigned rr=0;rr<_top_parts.size();rr++)
   {
      for(unsigned yy=0;yy<_ccs->constituents(_top_parts[rr]).size(); yy++)
      {
         top_constits.push_back(_alpha*_ccs->constituents(_top_parts[rr])[yy]);
      }
   }

   if(debugg)
   {
      cout << "number of top_constits: " << top_constits.size() << endl;
   }

   double filt_top_R = max(0.05, _drmin/2.0); // can be changed depending on resolution
   JetDefinition filtering_def(_jet_algorithm, filt_top_R);
   vector<PseudoJet> top_constits_filtered = Filtering(top_constits,filtering_def);
   PseudoJet topcandidate = Sum(top_constits_filtered);


   if(debugg) cout << "HPTTopTagger: topcandidate mass:" << topcandidate.m() << endl;

   if( topcandidate.m() < _mtmin || _mtmax < topcandidate.m() ) return;

   // obtain 3 subjets
   if(top_constits_filtered.size() < 3) return;
   JetDefinition reclustering(_jet_algorithm_recluster, 3.14/2);
   ClusterSequence cssubtop(top_constits_filtered,reclustering);
   vector <PseudoJet> top_subs = sorted_by_pt(cssubtop.exclusive_jets(3));
   _candjets.push_back(top_subs); //

   // transfer infos of the positively identified top to the outer world
   double deltatop = abs(topcandidate.m() - _mtmass);
   _delta_top = deltatop;
   _is_maybe_top = true;
   _top_candidate = topcandidate;
   store_topsubjets(top_subs);
   _top_hadrons=top_constits;
   /////////////////////// check mass plane cut////////////////////////
   _is_masscut_passed=check_mass_criteria(top_subs);
   _is_wmass_passed=check_Wmass(top_subs, _W_candidate);

   return;
}

vector<PseudoJet> HPTTopTagger::Filtering(const vector <PseudoJet> & top_constits, const JetDefinition & filtering_def)
{
   // perform filtering
   fastjet::ClusterSequence cstopfilt( top_constits, filtering_def);

   // extract top subjets
   vector<PseudoJet> filt_top_subjets = sorted_by_pt(cstopfilt.inclusive_jets());

   // take first n_topfilt subjets
   vector<PseudoJet> top_constits_filtered;
   for(unsigned ii = 0; ii<min(_nfilt, filt_top_subjets.size()) ; ii++)
   {
      cstopfilt.add_constituents(filt_top_subjets[ii],top_constits_filtered);
   }
   return top_constits_filtered;
}


PseudoJet HPTTopTagger::Sum(const vector<PseudoJet> & vec_pjet)
{
   PseudoJet sum;
   sum.reset(0.,0.,0.,0.);
   for(unsigned i=0;i<vec_pjet.size();i++){
      sum += vec_pjet.at(i);
   }
   return sum;
}

void HPTTopTagger::get_info() const
{
   cout << "maybe_top: " <<  _is_maybe_top << endl;
   cout << "mascut_passed: " <<  _is_masscut_passed << endl;
   cout << "top candidate mass:" <<  _top_candidate.m() << endl;
   cout << "top candidate (pt, eta, phi): ("
      <<  _top_candidate.perp() << ","
          <<  _top_candidate.eta() << ","
          <<  _top_candidate.phi_std() << ")" << endl;
   cout << "hadrons size: " <<  _top_hadrons.size() << endl;
   cout << "topcount: " <<  _top_count << endl;
   cout << "parts size: " <<  _parts_size << endl;
   cout << "delta_top: " <<  _delta_top << endl;
   return;
}


void HPTTopTagger::get_setting() const
{
   cout << "top mass: " <<  _mtmass << endl;
   cout << "W mass: " <<  _mwmass << endl;
   cout << "top mass range: [" << _mtmin << ", " << _mtmax << "]" << endl;
   cout << "W mass ratio range: [" << _rmin << ", " << _rmax << "] (["
      <<_rmin*_mtmass/_mwmass<< "%, "<< _rmax*_mtmass/_mwmass << "%])"<< endl;
   cout << "mass ratio cut: (m23cut, m13min, m13max)=("
      << _m23cut << ", " << _m13cutmin << ", " << _m13cutmax << ")" << endl;
   cout << "mass_drop_threshold: " << _mass_drop_threshold << endl;
   cout << "max_subjet_mass: " << _max_subjet_mass << endl;
   cout << "n_filtering: " << _nfilt << endl;
   cout << "JetAlgorithm for filtering: "<< _jet_algorithm << endl;
   cout << "JetAlgorithm for reclustering: "<< _jet_algorithm_recluster << endl;
   return;
}


#endif // __HPTTOPTAGGER_HH__
